//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// ---------------------------------------------------------------------------
//  Austin is a program for updating projects kept in XML files.
//
//  The parameters are:
//
//      [-?]            - Show usage and exit
//      [-v]            - Do validation
//		[-update]		- Update the times (Modified by -pessimistic)
//		[-pessimistic]	- The default (which this changes) is to do optimistic updates,
//							which assume the project is running smoothly.
//		[-png]			- Generate a png image of the Gantt chart.
//      filename        - The path to the project file to parse
//
//  These are case sensitive
// ---------------------------------------------------------------------------
#pragma warning(disable:4786)

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <string>
#include <iostream>
#include <algorithm>
#include <fstream>
using namespace std;
#include <util/PlatformUtils.hpp>
#include <parsers/DOMParser.hpp>
#include <dom/DOM_Node.hpp>
#include <dom/DOM_NamedNodeMap.hpp>
#include "DOMTreeErrorReporter.hpp"
#include "DirectedGraph.h"
#include "NodeIterator.h"
#include "Utility.h"
#include "Project.h"

const char * Version = "0.2.0";


// ---------------------------------------------------------------------------
//  Local data
//
//  xmlFile
//      The path to the file to parser. Set via command line.
//
//  doValidation
//      Indicates whether validation should be done. The default is not
//      to validate, but -v overrides that.
//
// ---------------------------------------------------------------------------
static char*	xmlFile         = 0;
static bool     doValidation    = false;


// ---------------------------------------------------------------------------
//  Forward references
// ---------------------------------------------------------------------------
void usage();

// ---------------------------------------------------------------------------
//
//  main
//
// ---------------------------------------------------------------------------
int main(int argC, char* argV[])
{
	bool update = false;
	bool optimistic = true;
#ifndef NO_PNG
	bool pngGantt = false;
#endif	
    // Initialize the XML4C2 system
    try
    {
        XMLPlatformUtils::Initialize();
    }

    catch(const XMLException& toCatch)
    {
        cerr << "Error during Xerces Initialization.\n"
             << "  Exception message:"
             << StrX(toCatch.getMessage()) << endl;
        return 1;
    }

    // Check command line and extract arguments.
    if (argC < 2)
    {
        usage();
        return 1;
    }

    // Watch for special case help request
    if (strcmp(argV[1], "-?") == 0)
    {
        usage();
        return 0;
    }

    // See if non validating dom parser configuration is requested.
    int parmInd;
    for (parmInd = 1; parmInd < argC; parmInd++)
    {
        // Break out on first parm not starting with a dash
        if (argV[parmInd][0] != '-')
            break;

        if (0 == strcmp(argV[parmInd], "-v")) {
            doValidation = true;
        } else if (0==strcmp(argV[parmInd], "-update")) {
            update = true;
        } else if (0==strcmp(argV[parmInd], "-pessimistic")) {
            optimistic = false;
#ifndef NO_PNG
        } else if (0==strcmp(argV[parmInd], "-png")) {
            pngGantt = true;
#endif
        } else {
            usage();
            return 1;
        }
    }

    //
    //  And now we have to have only one parameter left and it must be
    //  the file name.
    //
    if (parmInd + 1 != argC)
    {
        usage();
        return 1;
    }
    xmlFile = argV[parmInd];

    //
    //  Create our validator, then attach an error handler to the parser.
    //  The parser will call back to methods of the ErrorHandler if it
    //  discovers errors during the course of parsing the XML document.
    //
    DOMParser parser;
    ErrorHandler *errReporter = new DOMTreeErrorReporter();
    parser.setErrorHandler(errReporter);

    //
    //  Parse the XML file, catching any XML exceptions that might propogate
    //  out of it.
    //
    bool errorsOccured = false;
    try
    {
        parser.parse(xmlFile);
    }

    catch (const XMLException& e)
    {
        cerr << "An error occured during parsing\n   Message: "
             << StrX(e.getMessage()) << endl;
        errorsOccured = true;
    }
	// If the parse was successful, output the document data from the DOM tree
	if (!errorsOccured) {
		// Drill down to the project level 
		DOM_Node doc = parser.getDocument();
		Project project;
		DOM_Element projEle = (DOM_Element &)doc.getFirstChild();

		project.readConfig(projEle);
		project.rationalize();
		if (update) {	
			project.calculateStartAndFinishTimes(optimistic);
			project.writeConfig(projEle);

			ofstream outFile(xmlFile);

			if (!outFile) {
				cerr << "Could not open " << xmlFile << " for output.\n";
			} else {
				outFile << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
				//		outFile << "<!DOCTYPE project SYSTEM \"project.dtd\" >" << endl;
				outFile << (DOM_Node &)projEle << endl;		
				outFile.close();
			}
		}
		if (pngGantt) {
			generatePNG(project);
		}
	}		
	// Save out modified file.

    //
    //  Clean up the element counter object. The parser does not adopt handlers
    //  since they could be many objects or one object installed for multiple
    //  handlers.
    //
    delete errReporter;

    //
	//  The DOM document and its contents are reference counted, and need
	//	no explicit deletion.
    //
    return 0;
}


// ---------------------------------------------------------------------------
//
//  Usage()
//
// ---------------------------------------------------------------------------
void usage()
{
    cout << "Austin - Version " << Version << "\n"
		"Copyright (C) 2000 Joseph Gregorio\n"
		"Austin comes with ABSOLUTELY NO WARRANTY; see GPL.TXT for details.\n"
		"This is free software, and you are welcome to redistribute it\n"
		"under certain conditions as explained in GPL.TXT\n"
		"\n"
		    "\nUsage: Austin [options] file\n"
            "   This program updates a project named 'file'.\n"
            "   It will update the start and finish times of tasks.\n"
            "   Options are case sensitive\n\n"
            "   Options:\n"
            "     -v            Validate the data according to the DTD\n"
		    "     -update       Update the times (Modified by -pessimistic)\n"
            "     -pessimistic  The default (which this changes) is to do optimistic updates,\n"
            "                   which assume the project is running smoothly.\n"
            "     -png          Generate a png image of the Gantt chart.\n"
            "     -?            Show this help (must be the only parameter)\n"
          <<  endl;
}


















