//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef AUSTIN_DIRECTEDGRAPH_H
#define AUSTIN_DIRECTEDGRAPH_H

#include <set>
#include <map>
#include <vector>

// This graph template is used for the depth first search algorithm 
// below. It would need to be expanded for general use. The
// classes that can be put in here must support all the 
// operations needed to be put in the container set.
// This implementation assumes that all Vertices are unique.
// This implementation is based upon algorithms in "Introduction
// to Algorithms" by Cormem, Leiserson and Rivest.
template<class Vertex>
class DirectedGraph {
 private:
	// N.B. Has an entry for every single Vertex in the graph even if it has no edges.
	map<Vertex, set<Vertex> > adjacencyList_;  

 public:
	typedef Vertex value_type;
	// This is not used in the DirectedGraph class but later in some algorithms. Putting it
	// here avoids name clashes.
	enum Color {
		WHITE,
		GRAY,
		BLACK
	};
	
	DirectedGraph(){};

	~DirectedGraph(){};

	DirectedGraph(const DirectedGraph & rhs) {
		adjacencyList_ = rhs.adjacencyList_;
	};

	DirectedGraph & operator=(const DirectedGraph & rhs) {
		adjacencyList_ = rhs.adjacencyList_;
		return *this;
	}
	
	// Add a single Vertex.
	void push_back(const Vertex & vertex) {
		adjacencyList_.insert(make_pair(vertex, set<Vertex>()));
	}

	// Insert an Edge
	void insert_edge(const Vertex & begin, const Vertex & end) {
		adjacencyList_[begin].insert(end);
	}

	// Get the adjacency list for a Vertex.
	set<Vertex> adjacency(const Vertex & vertex) const {
		return adjacencyList_.find(vertex)->second;
	}

	// Get the predecessors for a Vertex.
	set<Vertex> predecessors(const Vertex & vertex) const {
		set<Vertex> returnValue;
		for(map<Vertex, set<Vertex> >::const_iterator i = adjacencyList_.begin(); i != adjacencyList_.end(); ++i) {
			if (i->second.find(vertex) != i->second.end()) {
				returnValue.insert(i->first);
			}
		}
		return returnValue;
	}

	// Get the set of all vertices in the graph.
	set<Vertex> vertices() const {
		set<Vertex> returnValue;
		for(map<Vertex, set<Vertex> >::const_iterator i = adjacencyList_.begin(); i != adjacencyList_.end(); ++i) {
			returnValue.insert(i->first);
		}
		return returnValue;
	}

	// Get the number of vertices in the graph.
	int size() const {
		return adjacencyList_.size();
	}

};

// Simple Depth First Search that pushes the results into the iterator given.
// Returns a list of vertices if a cycle is found.
template<class Vertex, class OutIt>
vector<Vertex> dfs(const DirectedGraph<Vertex> & graph, OutIt & outIter) {
	vector<Vertex> returnValue;
	map<Vertex, DirectedGraph<Vertex>::Color> color;
	set<Vertex> vertices = graph.vertices();
	for(set<Vertex>::iterator i = vertices.begin(); i != vertices.end(); ++i) {
		color[*i] = DirectedGraph<Vertex>::WHITE;		
	}
	for(i = vertices.begin(); i != vertices.end(); ++i) {
		if (DirectedGraph<Vertex>::WHITE == color[*i]) {
			vector<Vertex> cycle = dfs_visit(graph, *i, color, outIter);
			copy(cycle.begin(), cycle.end(), back_inserter(returnValue));
		}
	}	
	return returnValue;
}

// Utility function that should only be called from dfs().
// Returns a list of vertices if a cycle is found.
template<class Vertex, class OutIt>
vector<Vertex> dfs_visit(const DirectedGraph<Vertex> & graph, const Vertex & vertex, map<Vertex, DirectedGraph<Vertex>::Color> & color, OutIt & outIter) {
	vector<Vertex> returnValue;
	color[vertex] = DirectedGraph<Vertex>::GRAY;
	set<Vertex> adj = graph.adjacency(vertex);
	for(set<Vertex>::iterator i = adj.begin(); i != adj.end(); ++i) {
		if (DirectedGraph<Vertex>::WHITE == color[*i]) {
			vector<Vertex> cycle = dfs_visit(graph, *i, color, outIter);
			copy(cycle.begin(), cycle.end(), back_inserter(returnValue));
		} else if ((DirectedGraph<Vertex>::GRAY == color[*i]) && (vertex != *i)) {
			returnValue.push_back(*i);
			returnValue.push_back(vertex);			
		}
	}
	color[vertex] = DirectedGraph<Vertex>::BLACK;
	*outIter++ = vertex;
	return returnValue;
}


#endif

