//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#pragma warning(disable:4786)
#include "Task.h"
#include <cassert>
#include <sstream>
#include <dom/DOM_NodeList.hpp>
#include <dom/DOM_Document.hpp>
#include "Utility.h"
using namespace std;

class TaskImpl {
  public:
	TaskImpl();
	
	int count_;
	wstring name_;
	JulianDay start_;
	JulianDay finish_;
	JulianDay lateStart_;
	JulianDay lateFinish_;
	Duration work_;
	vector<pair<wstring, double> > resources_;
	double percentComplete_;
	vector<wstring> predecessors_;
};


TaskImpl::TaskImpl(){
	count_ = 1;
	percentComplete_ = 0.0f;
	resources_.push_back(make_pair(wstring(L"unknown"), 1.0));
}

Task::Task() {
	impl_ = new TaskImpl();
	
}

Task::Task(const Task & rhs) : impl_(rhs.impl_){
	impl_->count_++;
}

Task::~Task() {
	if (0 == --impl_->count_) {
		delete impl_;
	}
}

wstring	Task::name() const {
	return impl_->name_;
}

void	Task::name(const wstring & value) {
	impl_->name_ = value;
}

void 	Task::start(const JulianDay & value) {
	impl_->start_ = value;
}

// Eventually do this as a Julian date.
JulianDay 	Task::start() {
	return impl_->start_;
}

void 	Task::finish(const JulianDay & value) {
	impl_->finish_ = value;
}

JulianDay 	Task::finish() {
	return impl_->finish_;
}

void 	Task::lateStart(const JulianDay & value) {
	impl_->lateStart_ = value;
}

JulianDay 	Task::lateStart() {
	return impl_->lateStart_;
}

void 	Task::lateFinish(const JulianDay & value) {
	impl_->lateFinish_ = value;
}

JulianDay 	Task::lateFinish() {
	return impl_->lateFinish_;
}

Duration		Task::work() {
	return impl_->work_;
}

void    Task::work(const Duration & value) {
	impl_->work_ = value;
}

Duration 	Task::duration() {
	Duration returnValue = impl_->work_; // Just do this to get the right units.
	returnValue.set(work()/totalResources()+0.5);
	return returnValue;
}

void 		Task::resources(const vector<pair<wstring, double> > & value) {
	impl_->resources_ = value;
}

struct SumResources {
	double value_;
	SumResources() : value_(0) {}
	void operator()(pair<wstring, double> & p) {
		value_ += p.second;
	}
};

double 		Task::totalResources() {
	SumResources sum;
	sum = for_each(impl_->resources_.begin(), impl_->resources_.end(), sum);
	return sum.value_;
}

vector<pair<wstring, double> > 	Task::resources() {
	return impl_->resources_;
}

double 	Task::percentComplete() {
	return impl_->percentComplete_;
}

void 	Task::percentComplete(double value) {
	impl_->percentComplete_ = value;
}

vector<wstring>	& Task::predecessors() {
	return impl_->predecessors_;
}

Task & Task::operator=(const Task & rhs) {
	rhs.impl_->count_++;
	if (0 == --impl_->count_) {
		delete impl_;
	}
	impl_ = rhs.impl_;

	return *this;
}

bool operator<(const Task & lhs, const Task & rhs) {
	return (lhs.name() < rhs.name());
}

bool operator==(const Task & lhs, const Task & rhs) {
	return (lhs.name() == rhs.name());
}

bool operator!=(const Task & lhs, const Task & rhs) {
	return (lhs.name() != rhs.name());
}

struct ElementToResource {
	pair<wstring, double> operator()(DOM_Element resourceEle) {
		wstring name = (resourceEle.getAttribute(L"name").rawBuffer() ? resourceEle.getAttribute(L"name").rawBuffer() : L"");
		return make_pair<wstring, double>(name, wcstod(resourceEle.getAttribute(L"percent").rawBuffer(), NULL)/100);
	}
};

void Task::readConfig(DOM_Element taskElement) {
	impl_->name_ = taskElement.getAttribute(L"id").rawBuffer();

	impl_->work_.set(taskElement.getAttribute(L"work").rawBuffer());
	impl_->percentComplete_ 	= (double)getAttributeAsInt(taskElement, L"percent_complete", 0)/100.0f;	

	impl_->predecessors_.clear();
	vector<DOM_Element> predecessors = collectElements(L"predecessor", taskElement);
	transform(predecessors.begin(), predecessors.end(), back_inserter(impl_->predecessors_), ElementToName());

	impl_->resources_.clear();
	vector<DOM_Element> resources = collectElements(L"resource", taskElement);
	if (resources.size() > 0) {
		transform(resources.begin(), resources.end(), back_inserter(impl_->resources_), ElementToResource());
	} else {
		impl_->resources_.push_back(make_pair(wstring(L"unknown"), 1.0));
	}

	DOM_Element dates = getOrCreateByTagName(taskElement, L"dates");

	if (NULL != dates.getAttribute(L"start").rawBuffer()) {
		impl_->start_.set(dates.getAttribute(L"start").rawBuffer());
	}
	if (NULL != dates.getAttribute(L"finish").rawBuffer()) {
		impl_->finish_.set(dates.getAttribute(L"finish").rawBuffer());
	}
	impl_->lateStart_ = impl_->start_;
	impl_->lateFinish_ = impl_->finish_;
}

struct InsertPredecessor {
	DOM_Element taskElement_;
	InsertPredecessor(DOM_Element taskElement) : taskElement_(taskElement) {}
	void operator()(wstring & name) {
		DOM_Element predNode = taskElement_.getOwnerDocument().createElement(L"predecessor");
		predNode.setAttribute(L"id", name.c_str());
		taskElement_.appendChild(predNode);
	}
};

struct RemovePredecessor {
	DOM_Element taskElement_;
	RemovePredecessor(DOM_Element taskElement) : taskElement_(taskElement) {}
	void operator()(wstring & name) {
		vector<DOM_Element> predecessors = collectElements(L"predecessor", taskElement_);
		vector<DOM_Element>::iterator predNodeIt = find_if(predecessors.begin(), predecessors.end(), HasID(name.c_str()));
		if (predecessors.end() != predNodeIt) {
			taskElement_.removeChild(*predNodeIt);
		}
	}
};

struct InsertResource {
	DOM_Element taskElement_;
	InsertResource(DOM_Element taskElement) : taskElement_(taskElement) {}
	void operator()(pair<wstring, double> & resource) {
		DOM_Element predNode = taskElement_.getOwnerDocument().createElement(L"resource");
		predNode.setAttribute(L"name", resource.first.c_str());
		wstringstream s;
		s << (int)(resource.second*100+0.5);
		predNode.setAttribute(L"percent", s.str().c_str());
		taskElement_.appendChild(predNode);
	}
};

void Task::writeConfig(DOM_Element taskElement) {
	taskElement.setAttribute(L"id", impl_->name_.c_str());

	DOM_Element dates = getOrCreateByTagName(taskElement, L"dates");

	dates.setAttribute(L"start", 	impl_->start_.asString().c_str());
	dates.setAttribute(L"finish", 	impl_->finish_.asString().c_str());

	setAttributeWithInt(taskElement, L"percent_complete", impl_->percentComplete_*100 + 0.5);	
	taskElement.setAttribute(L"work", impl_->work_.get().c_str());
	

	// Update the list of predecessors. Add in the ones added and remove the ones we have dropped.

	vector<DOM_Element> orig_predecessors = collectElements(L"predecessor", taskElement);
	vector<wstring> orig_pred_names;
	transform(orig_predecessors.begin(), orig_predecessors.end(), back_inserter(orig_pred_names), ElementToName());
	// We need to sort to do set_difference().
	sort(orig_pred_names.begin(), orig_pred_names.end());
	sort(impl_->predecessors_.begin(), impl_->predecessors_.end());
	vector<wstring> toBeRemoved;
	vector<wstring> toBeInserted;
	// Determine which predecessors have been added and removed.
	set_difference(orig_pred_names.begin(), orig_pred_names.end(), impl_->predecessors_.begin(), impl_->predecessors_.end(), back_inserter(toBeRemoved));
	set_difference(impl_->predecessors_.begin(), impl_->predecessors_.end(), orig_pred_names.begin(), orig_pred_names.end(), back_inserter(toBeInserted));
	// Now change the DOM to reflect the changes.
	for_each(toBeInserted.begin(), toBeInserted.end(), InsertPredecessor(taskElement));
	for_each(toBeRemoved.begin(), toBeRemoved.end(), RemovePredecessor(taskElement));	

	// Write out the resource, note we don't have to worry about updating them since we don't add or remove them like we do predecessors.
	vector<DOM_Element> orig_resources = collectElements(L"resource", taskElement);
	if (orig_resources.size() != impl_->resources_.size()) {
		for_each(impl_->resources_.begin(), impl_->resources_.end(), InsertResource(taskElement));
	}
	
}














