//    Austin is a program for updating projects kept in XML files.
//    Copyright (C) 2000  Joseph Gregorio
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#include "Utility.h"
#include "NodeIterator.h"
#include <util/PlatformUtils.hpp>
#include <dom/DOM_NodeList.hpp>
#include <dom/DOM_Document.hpp>
#include <dom/DOM_NamedNodeMap.hpp>
#include <sstream>
#include <algorithm>
#include <cctype>
#include "Project.h"
#include <ctime>
#include <cwchar>
using namespace std;

#ifndef NO_PNG
/* Bring in gd library functions */
#include "gd.h"
#include "gdfontmb.h"
#include "gdfontt.h"
#endif

using namespace std;


void outputNode(ostream& target, const DOM_Node & toWrite, int level);

// ---------------------------------------------------------------------------
//  doEscapes
//      Indicates whether special chars should be escaped in the output.
// ---------------------------------------------------------------------------
static bool     doEscapes       = true;

wstring stripLeadingTrailingWhiteSpace(const wstring & s) {
	wstring::const_iterator begin = s.begin();
	wstring::const_iterator end = s.end()-1;
	while (isspace(*begin)) {
		++begin;
	}
	while (isspace(*end)) {
		--end;
	}
	return wstring(begin, end+1);
}


string  stringMBCS(const wchar_t * value) {
	unsigned int mbcsSize = ::wcstombs(0, value, 0);
	char * buffer = new char[mbcsSize+1];
	wcstombs(buffer, value, mbcsSize);
	buffer[mbcsSize] = 0;
	string returnValue = buffer;
	delete [] buffer;
	return returnValue;	
}

// ---------------------------------------------------------------------------
//
//  collectElements -
//		Given a DOM_Node this will pull out all the children 
//		that are DOM_Elements with the given name.        
//
// ---------------------------------------------------------------------------
vector<DOM_Element> collectElements(const wstring & name, DOM_Node parent) {
	vector<DOM_Element> nodes;
	copy(NodeIterator<DOM_Element> ((DOM_Element &)parent.getFirstChild()), NodeIterator<DOM_Element> (), back_inserter(nodes));
	vector<DOM_Element>::iterator endTasks = partition(nodes.begin(), nodes.end(), HasName(name));
	nodes.erase(endTasks, nodes.end());
	
	return nodes;
}

// ---------------------------------------------------------------------------
//
//  getSingleElementByName -
//		Get a single element child of a DOM_Element by element name.
//
// ---------------------------------------------------------------------------
DOM_Element getSingleElementByName(DOM_Element element, const wstring & name) {
		return (DOM_Element &) element.getElementsByTagName(name.c_str()).item(0);
}

int getAttributeAsInt(DOM_Element element, const wstring & name, int defaultValue) {
	int returnValue = defaultValue;

	wstring value;
	if (!element.getAttributeNode(name.c_str()).isNull()) {
		value = element.getAttribute(name.c_str()).rawBuffer();
	}
	if (L"" != value) {
		wstringstream s(value);
		s >> returnValue;
	}

	return returnValue;
}

void setAttributeWithInt(DOM_Element element, const wstring & name, int value) {
	wstringstream s;
	s << value;
	element.setAttribute(name.c_str(), s.str().c_str());
}


DOM_Element getOrCreateByTagName(DOM_Element element, const wstring & tagName) {
	DOM_NodeList nodeList = element.getElementsByTagName(tagName.c_str());
	DOM_Element returnValue;
	if (nodeList.getLength() < 1) {
		returnValue = element.getOwnerDocument().createElement(tagName.c_str());
		element.appendChild(returnValue);
	} else {
		returnValue = (DOM_Element &) nodeList.item(0);
	}

	return returnValue;
}

// ---------------------------------------------------------------------------
//
//  outputContent  - Write document content from a DOMString to a C++ ostream.
//                   Escape the XML special characters (<, &, etc.) unless this
//                   is suppressed by the command line option.
//
// ---------------------------------------------------------------------------
void outputContent(ostream& target, const DOMString &toWrite)
{
    
    if (doEscapes == false)
    {
        target << toWrite;
    }
     else
    {
        int            length = toWrite.length();
        const XMLCh*   chars  = toWrite.rawBuffer();
        
        int index;
        for (index = 0; index < length; index++)
        {
            switch (chars[index])
            {
            case chAmpersand :
                target << "&amp;";
                break;
                
            case chOpenAngle :
                target << "&lt;";
                break;
                
            case chCloseAngle:
                target << "&gt;";
                break;
                
            case chDoubleQuote :
                target << "&quot;";
                break;
                
            default:
                // If it is none of the special characters, print it as such
                target << StrX(&chars[index], 1);
                break;
            }
        }
    }

    return;
}


// ---------------------------------------------------------------------------
//
//  ostream << DOMString      Stream out a DOM string.
//
// ---------------------------------------------------------------------------
ostream& operator<<(ostream& target, const DOMString& s)
{
    if (s.length() > 0)
        target << StrX(s.rawBuffer(), s.length());
    return target;
}
// ---------------------------------------------------------------------------
//
//  ostream << DOM_Node   
//
//                Stream out a DOM node, and, recursively, all of its children.
//                This function is the heart of writing a DOM tree out as
//                XML source.  Give it a document node and it will do the whole thing.
//
// ---------------------------------------------------------------------------
ostream& operator<<(ostream& target, const DOM_Node & toWrite) {
	outputNode(target, toWrite, 0);
	return target;
}

void outputNode(ostream& target, const DOM_Node & toWrite, int level)
{
    // Get the name and value out for convenience
    DOMString   nodeName = toWrite.getNodeName();
    DOMString   nodeValue = toWrite.getNodeValue();

	switch (toWrite.getNodeType())
    {
		case DOM_Node::TEXT_NODE:
        {
			wstring v(nodeValue.rawBuffer());
			wstring::iterator begin = find_if(v.begin(), v.end(), isalnum);
			wstring::iterator end = v.end()-1;
			while (isspace(*end)) {	
				--end;
			}			
			if (end-begin > 0) {
				wstring stripped(begin, end+1);
				outputContent(target, stripped.c_str());
			}
            break;
        }

        case DOM_Node::PROCESSING_INSTRUCTION_NODE :
        {
            target  << "<?"
                    << nodeName
                    << ' '
                    << nodeValue
                    << "?>";
            break;
        }

        case DOM_Node::DOCUMENT_NODE :
        {
            target << "<?xml version='1.0' encoding='utf-8' ?>\n";
            DOM_Node child = toWrite.getFirstChild();
            while( child != 0)
            {
				outputNode(target, child, level+1);
                child = child.getNextSibling();
            }

            break;
        }

        case DOM_Node::ELEMENT_NODE :
        {
            // Output the element start tag.
			target << endl;
			fill_n(ostream_iterator<char>(target), level, '\t');
			
            target << '<' << nodeName;

            // Output any attributes on this element
            DOM_NamedNodeMap attributes = toWrite.getAttributes();
            int attrCount = attributes.getLength();
            for (int i = 0; i < attrCount; i++)
            {
                DOM_Node  attribute = attributes.item(i);

                target  << ' ' << attribute.getNodeName()
                        << " = \"";
                        //  Note that "<" must be escaped in attribute values.
                        outputContent(target, attribute.getNodeValue());
                        target << '"';
            }

            //
            //  Test for the presence of children, which includes both
            //  text content and nested elements.
            //
            DOM_Node child = toWrite.getFirstChild();
            if (child != 0)
            {
                // There are children. Close start-tag, and output children.
                target << ">";
                while( child != 0)
                {
					outputNode(target, child, level+1);
                    child = child.getNextSibling();
                }

                // Done with children.  Output the end tag.
				target << endl;
				fill_n(ostream_iterator<char>(target), level, '\t');
                target << "</" << nodeName << ">";
            }
            else
            {
                //
                //  There were no children.  Output the short form close of the
                //  element start tag, making it an empty-element tag.
                //
                target << "/>";
            }
            break;
        }

        case DOM_Node::ENTITY_REFERENCE_NODE:
        {
            DOM_Node child;
            for (child = toWrite.getFirstChild(); child != 0; child = child.getNextSibling()) {
				outputNode(target, child, level+1);
			}
            break;
        }

        case DOM_Node::CDATA_SECTION_NODE:
        {
            target << "<![CDATA[" << nodeValue << "]]>";
            break;
        }

        case DOM_Node::COMMENT_NODE:
        {
            target << "<!--" << nodeValue << "-->";
            break;
        }

        default:
            cerr << "Unrecognized node type = "
                 << (long)toWrite.getNodeType() << endl;
    }
}

// ---------------------------------------------------------------------------
//  StrX: Private helper methods
// ---------------------------------------------------------------------------
void StrX::transcode(const XMLCh* const toTranscode, const unsigned int len)
{
    // Short circuit if its a null pointer
    if (!toTranscode || (!toTranscode[0]))
    {
        fLocalForm = new char[1];
        fLocalForm[0] = 0;
        return;
	}

    // See if our XMLCh and wchar_t as the same on this platform
    const bool isSameSize = (sizeof(XMLCh) == sizeof(wchar_t));

    //
    //  Get the actual number of chars. If the passed len is zero, its null
    //  terminated. Else we have to use the len.
    //
    wchar_t realLen = (wchar_t)len;
    if (!realLen)
    {
        //
        //  We cannot just assume we can use wcslen() because we don't know
        //  if our XMLCh is the same as wchar_t on this platform.
        //
        const XMLCh* tmpPtr = toTranscode;
        while (*(tmpPtr++))
            realLen++;
    }

    //
    //  If either the passed length was non-zero or our char sizes are not 
    //  same, we have to use a temp buffer. Since this is common in these
    //  samples, we just do it anyway.
    //
    wchar_t* tmpSource = new wchar_t[realLen + 1];
    if (isSameSize)
    {
        memcpy(tmpSource, toTranscode, realLen * sizeof(wchar_t));
    }
     else
    {
        for (unsigned int index = 0; index < realLen; index++)
            tmpSource[index] = (wchar_t)toTranscode[index];
    }
    tmpSource[realLen] = 0;

    // See now many chars we need to transcode this guy
    const unsigned int targetLen = ::wcstombs(0, tmpSource, 0);

    // Allocate out storage member
    fLocalForm = new char[targetLen + 1];

    //
    //  And transcode our temp source buffer to the local buffer. Cap it
    //  off since the converter won't do it (because the null is beyond
    //  where the target will fill up.)
    //
    ::wcstombs(fLocalForm, tmpSource, targetLen);
    fLocalForm[targetLen] = 0;

    // Don't forget to delete our temp buffer
    delete [] tmpSource;
}

#ifndef NO_PNG

void generatePNG(Project & project) {

	const int IMAGE_WIDTH = 800;

	gdFontPtr font = gdFontMediumBold;
	gdFontPtr smfont = gdFontTiny;
	int LINE_HEIGHT = font->h + 5; // 2 spaces above, 2 spaces below and a seperating line.
	int MAX_CHARS_TASK_NAME = 20;
	int MAX_CHARS_TASK_DURATION = 5;
	int NAME_WIDTH  = font->w * MAX_CHARS_TASK_NAME;
	int DURATION_WIDTH  = font->w * MAX_CHARS_TASK_DURATION;
	int NON_BAR_WIDTH = NAME_WIDTH + DURATION_WIDTH;
	int TASK_BAR_AREA_WIDTH = IMAGE_WIDTH - NON_BAR_WIDTH;
	
	/* Declare the image */
	gdImagePtr im;
	/* Declare an output file */
	FILE *out;

	vector<Task> tasks = project.tasksPartiallyOrdered();
	int numberOfTasks = tasks.size();
	
	int IMAGE_HEIGHT = (numberOfTasks+2)*LINE_HEIGHT;
	im = gdImageCreate(IMAGE_WIDTH, IMAGE_HEIGHT); // Leave space at the top for the title.

	/* Allocate the color indices.
		Since this is the first color in a new image, it will
		be the background color. */
	int backgroundColor = gdImageColorAllocate(im, 245, 245, 255);
	//	int backgroundColor = gdImageColorAllocate(im, 100, 101, 114);
	int gridlinesColor = gdImageColorAllocate(im, 128, 128, 128);
	int taskTextColor = gdImageColorAllocate(im, 0, 0, 0);
	int taskWorkRectColor = gdImageColorAllocate(im, 0, 0, 128);
	int taskCompletedColor = gdImageColorAllocate(im, 0, 0, 255);
	int titleColor = gdImageColorAllocate(im, 128, 0, 0);
	int todayColor = gdImageColorAllocate(im, 192, 0, 0);

	// Find the earliest start and latest finish times.
	// Now scale all lines to that scale.
	JulianDay start = project.find(L"Start").start();
	JulianDay finish = project.find(L"Finish").finish();
	Duration projectDuration;
	projectDuration.set(finish - start);

	wstring projectName = stripLeadingTrailingWhiteSpace(project.name());
	gdImageString(im, font, 1, 1,(unsigned char *)StrX(projectName.c_str(), projectName.size()).localForm(), titleColor);

	gdImageLine(im, NON_BAR_WIDTH, LINE_HEIGHT, IMAGE_WIDTH-1, LINE_HEIGHT, gridlinesColor);	

	int yOffset = 2*LINE_HEIGHT;
	for (int i=0; i<numberOfTasks; i++) {
		yOffset = 2*LINE_HEIGHT + i*LINE_HEIGHT;
		gdImageLine(im, 0, yOffset, IMAGE_WIDTH-1, yOffset, gridlinesColor);	
		Duration workBegins;
		workBegins.set(tasks[i].start() - start);

		double workPercent       	= ((double)(int)(tasks[i].finish() - tasks[i].start()))/((double)(int)projectDuration);
		double percentPercent       = (( tasks[i].percentComplete() * (double)(int)(tasks[i].finish() - tasks[i].start())) / (double)(int)projectDuration);
		double workBeginsPercent 	= ((double)(int)workBegins/(double)(int)projectDuration);
		int workXLeft 				= workBeginsPercent*TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;
		int workXRight 				= (workBeginsPercent + workPercent)*TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;
		int percentCompleteXRight 	= (workBeginsPercent + percentPercent)*TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;

		gdImageFilledRectangle(im, workXLeft, yOffset+1, percentCompleteXRight, yOffset+LINE_HEIGHT-1, taskCompletedColor);
		gdImageRectangle(im, workXLeft, yOffset+1, workXRight, yOffset+LINE_HEIGHT-1, taskWorkRectColor);
		gdImageRectangle(im, workXLeft+1, yOffset+2, workXRight-1, yOffset+LINE_HEIGHT-2, taskWorkRectColor);
		wstring taskName = tasks[i].name();
		gdImageString(im, font, 2, yOffset+1, (unsigned char *)StrX(taskName.c_str(), taskName.size()).localForm(), taskTextColor);

		wstring taskShortDuration = tasks[i].duration().getShortString();
		gdImageString(im, font, NAME_WIDTH + (MAX_CHARS_TASK_DURATION - taskShortDuration.size()) * smfont->w, yOffset+1, (unsigned char *)StrX(taskShortDuration.c_str(), taskShortDuration.size()).localForm(), taskTextColor);
	}
	gdImageLine(im, NON_BAR_WIDTH-1, 0, NON_BAR_WIDTH-1, IMAGE_HEIGHT, gridlinesColor);	
	gdImageLine(im, NAME_WIDTH-1, 0, NAME_WIDTH-1, IMAGE_HEIGHT, gridlinesColor);	


	// Draw date marks 
	JulianDay blockStart = start;
	struct tm startTm = start.asTm();
	
	enum BlockLength {
		DAY,
		WEEK,
		MONTH,
		YEAR
	};
	BlockLength typeOfBlock;
	char * dateFormat[] = {
		"%m/%d",
		"%m/%d",
		"%m/%Y",
		"%m/%Y",
	};	
	
	if (projectDuration <= 10) {
		typeOfBlock = DAY;
	} else if (projectDuration / 7 <= 26) {
		startTm.tm_mday += (8-startTm.tm_wday)%7;
		mktime(&startTm);
		typeOfBlock = WEEK;
	} else if (projectDuration / 30 <= 16) {
		if (1 != startTm.tm_mday) {
			startTm.tm_mday = 1;
			startTm.tm_mon += 1;
			mktime(&startTm);
		}
		typeOfBlock = MONTH;
	} else {
		if (0 != startTm.tm_mon || 1 == startTm.tm_mday) {
			startTm.tm_mon = 0;
			startTm.tm_mday = 1;
			startTm.tm_year += 1;
			mktime(&startTm);			
		}
		typeOfBlock = YEAR;
	}
	if (projectDuration > 0) {
		blockStart.set(&startTm);
		int blockLeft = ((double)(blockStart - start)/(double)(int)projectDuration) * TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;
		while (blockLeft < IMAGE_WIDTH) {
			gdImageLine(im, blockLeft, 0, blockLeft, LINE_HEIGHT, gridlinesColor);

			char s[15];
			strftime(s, 15, dateFormat[typeOfBlock], &startTm);
			gdImageString(im, smfont, blockLeft+1, 1, (unsigned char *)s, taskTextColor);
			startTm = blockStart.asTm();
			switch (typeOfBlock) {
			case DAY:
				startTm.tm_mday += 1;
				break;
			case WEEK:
				startTm.tm_mday += 7;
				break;
			case MONTH:
				startTm.tm_mon += 1;
				break;
			case YEAR:
				startTm.tm_year += 1;
				break;
			}
			mktime(&startTm);
			blockStart.set(&startTm);
			blockLeft = ((double)(blockStart - start)/(double)(int)projectDuration) * TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;
			if (0 == (blockStart - start)) {
				break;  // This can happen if you run Austin on an initial project file and don't specify -update.
			}
		}
	}
	// Draw a line at today.
	JulianDay today;
	if (today - start > 0 && finish-today > 0) {
		int todayLeft = ((double)(today - start)/(double)(int)projectDuration) * TASK_BAR_AREA_WIDTH + NON_BAR_WIDTH;
		gdImageLine(im, todayLeft, 2*LINE_HEIGHT, todayLeft, IMAGE_HEIGHT, todayColor);
	}
	

	/* Open a file for writing. "wb" means "write binary", important
		under MSDOS, harmless under Unix. */
	wstring pngFileName = projectName + L".png";
	unlink(stringMBCS(pngFileName.c_str()).c_str());
	out = fopen(stringMBCS(pngFileName.c_str()).c_str(), "wb");
	
	if (out) {
		/* Output the image to the disk file. */
		gdImagePng(im, out);	

		/* Close the file. */
		fclose(out);

		/* Destroy the image in memory. */
		gdImageDestroy(im);
	} else {
		wcerr << L"Could not create image file: " << pngFileName << endl;
	}
}
#endif

